# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from typing import Dict

from .base import fire_opal_workflow


@fire_opal_workflow("ibm_mitigation_workflow")
def create_mitigation_data(
    ibm_device_name: str,
    ibmq_token: str,
    hub: str,
    group: str,
    project: str,
) -> Dict:
    """
    Creates mitigation data for Fire Opal.

    Parameters
    ----------
    ibm_device_name : str
        The device to run mitigation on.
    ibmq_token : str
        The IBM token for authentication.
    hub : str
        The IBM hub for authentication.
    group : str
        The IBM group for authentication.
    project : str
        The IBM project for authentication.

    Returns
    -------
    Dict
        Dictionary with workflow output.
    """

    return {
        "ibm_device_name": ibm_device_name,
        "ibmq_token": ibmq_token,
        "hub": hub,
        "group": group,
        "project": project,
    }



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from __future__ import annotations

from qctrlclient.core import print_warnings
from qctrlcommons.preconditions import check_argument

from fireopal.config import get_config
from fireopal.credentials import Credentials

from .base import fire_opal_workflow


@fire_opal_workflow("compile_and_run_workflow", formatter=print_warnings)
def execute(
    circuits: str | list[str],
    shot_count: int,
    credentials: Credentials,
    backend_name: str,
) -> dict:
    """
    Execute a batch of `circuits` where `shot_count` measurements are taken per circuit.

    Parameters
    ----------
    circuits : str or list[str]
        Quantum circuit(s) in the form of a QASM strings. You may use Qiskit to
        generate these strings.
    shot_count : int
        Number of bitstrings that are sampled from the final quantum state.
    credentials : Credentials
        The credentials for running circuits. See the `credentials` module for functions
        to generate credentials for your desired provider.
    backend_name : str
        The backend device name that should be used to run circuits.

    Returns
    -------
    dict
        A dictionary containing probability mass functions and warnings.
    """
    _check_execute_types(
        circuits=circuits,
        shot_count=shot_count,
        credentials=credentials,
        backend_name=backend_name,
    )
    circuits = _handle_single_circuit(circuits)

    settings = get_config()
    credentials_with_org = credentials.copy()
    credentials_with_org.update({"organization": settings.organization})
    return {
        "circuits": circuits,
        "shot_count": shot_count,
        "credentials": credentials_with_org,
        "backend_name": backend_name,
    }


def _handle_single_circuit(strings: str | list[str]) -> list[str]:
    """
    Convert a single string to a list holding containing it, if applicable.

    Parameters
    ----------
    strings : str or list[str]
        One or more strings.

    Returns
    -------
    list[str]
        The input if originally provided as a list. Otherwise, the input string in
        a list.
    """
    if isinstance(strings, str):
        return [strings]
    return strings


def _check_execute_types(
    circuits: str | list[str],
    shot_count: int,
    credentials: Credentials,
    backend_name: str,
) -> None:
    """
    Check if the types are correct for execute function.

    Parameters
    ----------
    circuits : str or list[str]
        Quantum circuit(s) in the form of a QASM strings. You may use Qiskit to
        generate these strings.
    shot_count : int
        Number of bitstrings that are sampled from the final quantum state.
    credentials : Credentials
        The credentials for running circuits. See the `credentials` module for functions
        to generate credentials for your desired provider.
    backend_name : str
        The backend device name that should be used to run circuits.
    """

    if isinstance(circuits, list):
        check_argument(
            all(isinstance(circuit, str) for circuit in circuits),
            "Invalid type received for circuits input. All circuits must be strings.",
            {"circuits": circuits},
        )
    else:
        check_argument(
            isinstance(circuits, str),
            "Invalid type received for circuits input. The circuit must be a string.",
            {"circuits": circuits},
        )
    check_argument(
        isinstance(shot_count, int),
        "Invalid type received for shot_count input. The shot_count must be an integer.",
        {"shot_count": shot_count},
    )
    check_argument(
        isinstance(credentials, dict),
        "Invalid type received for credentials input. The credentials must be a dictionary.",
        {
            "credentials": credentials,
            "credentials type": type(credentials),
        },
    )
    check_argument(
        isinstance(backend_name, str),
        "Invalid type received for backend_name input. The backend_name must be a string.",
        {"backend_name": backend_name},
    )



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from typing import (
    Any,
    Dict,
)

from .base import fire_opal_workflow


@fire_opal_workflow("benchmark_workflow")
def benchmark(
    ibm_device_name: str,
    benchmark_config: Dict[str, Any],
    hub: str,
    group: str,
    project: str,
    ibmq_token: str,
) -> Dict:
    """
    Runs the Fire Opal benchmark suite.

    Parameters
    ----------
    ibm_device_name : str
        Description.
    benchmark_config : Dict[str, Any]
        Description.
    hub : str
        Description.
    group : str
        Description.
    project : str
        Description.
    ibmq_token : str
        Description.

    Returns
    -------
    Dict
        The output of the benchmarking workflow.
    """
    return {
        "ibm_device_name": ibm_device_name,
        "benchmark_config": benchmark_config,
        "hub": hub,
        "group": group,
        "project": project,
        "ibmq_token": ibmq_token,
    }



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

"""
Module to host some utility functions, which are exposed at the top-level.
"""
import sys
from pathlib import Path
from types import ModuleType

import tomli
from importlib_metadata import packages_distributions

__all__ = ["print_package_versions"]


def print_package_versions() -> None:
    """
    Print a Markdown-formatted table showing the Python version being used,
    as well as the versions of some loaded packages that are relevant to the Fire Opal
    client package.
    """

    # packages_distributions returns a map between the top-level module names and package names.
    # However, it doesn't understand packages installed in editable mode,
    # which are handled in _get_package_name.
    _package_names_mapping = packages_distributions()

    def _get_package_name(module: ModuleType) -> str:
        if module.__name__ in _package_names_mapping:
            return _package_names_mapping[module.__name__][0]

        # The package is in editable mode: look in pyproject.toml to get the package name.
        with open(
            Path.joinpath(Path(module.__path__[0]).parent, "pyproject.toml"), "rb"
        ) as file:
            config = tomli.load(file)
            return config["tool"]["poetry"]["name"]

    top_level_module_names = [
        # External packages.
        "matplotlib",
        "networkx",
        "numpy",
        "qiskit",
        "qiskit-ibm-provider",
        "sympy",
        # Q-CTRL packages.
        "fireopal",
        "qctrlvisualizer",
    ]

    python_version = (
        f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
    )

    # List containing the items in the different rows.
    table_items = [
        (
            _get_package_name(sys.modules[module_name]),
            sys.modules[module_name].__version__,
        )
        for module_name in top_level_module_names
        if module_name in sys.modules
    ]

    # Widths of the table columns.
    package_width = max(max(len(item[0]) for item in table_items), 7)
    version_width = max(max(len(item[1]) for item in table_items), 7)

    # Add headers and Python version at top of table.
    table_items = [
        ("Package", "Version"),
        ("-" * package_width, "-" * version_width),
        ("Python", python_version),
    ] + table_items

    # Print table.
    for name, version_ in table_items:
        print(f"| {name:{package_width}s} | {version_:{version_width}s} |")
